From 5f94c9d780fa4e789dbbd992da80ae7daf40e79f Mon Sep 17 00:00:00 2001
From: Zhigang Wang <pop.felix@gmail.com>
Date: Thu, 2 Apr 2009 23:15:33 +0800
Subject: [PATCH] GSoC 09 qualification task: Add Truran VOD access module

---
 configure.ac                |   16 +-
 modules/access/Modules.am   |    1 +
 modules/access/truran_vod.c |  604 +++++++++++++++++++++++++++++++++++++++++++
 3 files changed, 618 insertions(+), 3 deletions(-)
 create mode 100644 modules/access/truran_vod.c

diff --git a/configure.ac b/configure.ac
index ffc4d36..241da16 100644
--- a/configure.ac
+++ b/configure.ac
@@ -348,11 +348,11 @@ case "${host_os}" in
         VLC_ADD_LDFLAGS([vlc],[-mwindows])
         VLC_ADD_LIBS([activex mozilla],[-lgdi32])
         VLC_ADD_LIBS([cdda vcdx cddax sdl_image],[-lwinmm])
-        VLC_ADD_LIBS([access_http access_mms access_udp access_tcp access_ftp access_rtmp access_output_udp access_output_shout access_output_rtmp sap slp http stream_out_standard stream_out_rtp stream_out_raop vod_rtsp access_realrtsp rtp telnet rc netsync gnutls growl_udp flac ts audioscrobbler lua remoteosd],[-lws2_32])
+        VLC_ADD_LIBS([access_http access_truran_vod access_mms access_udp access_tcp access_ftp access_rtmp access_output_udp access_output_shout access_output_rtmp sap slp http stream_out_standard stream_out_rtp stream_out_raop vod_rtsp access_realrtsp rtp telnet rc netsync gnutls growl_udp flac ts audioscrobbler lua remoteosd],[-lws2_32])
     fi
     if test "${SYS}" = "mingwce"; then
         # add ws2 for closesocket, select, recv
-        VLC_ADD_LIBS([libvlc access_http access_mms access_udp access_tcp access_ftp access_rtmp access_output_udp sap http netsync audioscrobbler growl rtp stream_out_rtp remoteosd],[-lws2])
+        VLC_ADD_LIBS([libvlc access_http access_truran_vod access_mms access_udp access_tcp access_ftp access_rtmp access_output_udp sap http netsync audioscrobbler growl rtp stream_out_rtp remoteosd],[-lws2])
         VLC_ADD_LIBS([libvlc],[-lmmtimer])
    fi
     ;;
@@ -553,7 +553,7 @@ AC_CHECK_FUNCS(connect,,[
 
 AC_CHECK_FUNCS(send,,[
   AC_CHECK_LIB(socket,send,[
-    VLC_ADD_LIBS([access_http access_mms access_udp access_tcp access_ftp access_rtmp sap access_output_udp access_output_rtmp stream_out_standard growl_udp remoteosd],[-lsocket])
+    VLC_ADD_LIBS([access_http access_truran_vod access_mms access_udp access_tcp access_ftp access_rtmp sap access_output_udp access_output_rtmp stream_out_standard growl_udp remoteosd],[-lsocket])
   ])
 ])
 
@@ -2210,6 +2210,16 @@ then
 fi
 
 dnl
+dnl  Truran VOD access module
+dnl
+AC_ARG_ENABLE(truran,
+  [  --enable-truran         Truran VOD stream support (default disabled)])
+if test "${enable_truran}" = "yes"
+then
+  VLC_ADD_PLUGIN([access_truran_vod])
+fi
+
+dnl
 dnl  Windows DirectShow access module
 dnl
 AC_ARG_ENABLE(dshow,
diff --git a/modules/access/Modules.am b/modules/access/Modules.am
index d8fda93..bbf735a 100644
--- a/modules/access/Modules.am
+++ b/modules/access/Modules.am
@@ -19,6 +19,7 @@ SOURCES_access_dv = dv.c
 SOURCES_access_udp = udp.c
 SOURCES_access_tcp = tcp.c
 SOURCES_access_http = http.c
+SOURCES_access_truran_vod = truran_vod.c
 SOURCES_access_ftp = ftp.c
 SOURCES_access_smb = smb.c
 SOURCES_access_gnomevfs = gnomevfs.c
diff --git a/modules/access/truran_vod.c b/modules/access/truran_vod.c
new file mode 100644
index 0000000..61001dd
--- /dev/null
+++ b/modules/access/truran_vod.c
@@ -0,0 +1,604 @@
+/*****************************************************************************
+ * truran_vod.c : Truran VOD access module
+ *****************************************************************************
+ * Copyright (C) 2009 the VideoLAN team
+ * $Id$
+ *
+ * Authors:
+ *
+ * This program is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation; either version 2 of the License, or
+ * (at your option) any later version.
+ *
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, write to the Free Software
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
+ *****************************************************************************/
+
+/*****************************************************************************
+ * Preamble
+ *****************************************************************************/
+
+#ifdef HAVE_CONFIG_H
+# include "config.h"
+#endif
+
+#include <vlc_common.h>
+#include <vlc_plugin.h>
+#include <vlc_access.h>
+#include <vlc_network.h>
+#include <vlc_url.h>
+
+/*
+vod://202.120.2.242:1680/\\202.120.63.123\ftp3$\movie\test.avi
+#define VOD_DEBUG
+#define TRURAN_BLOCK_ACCESS
+*/
+
+#define MTU (32*1024)
+
+#ifdef VOD_DEBUG
+#include <stdio.h>
+#endif
+
+#ifdef HAVE_ATTRIBUTE_PACKED
+#define ATTRIBUTE_PACKED __attribute__ ((packed))
+#define PRAGMA_PACK 0
+#else
+#define ATTRIBUTE_PACKED
+#define PRAGMA_PACK 1
+#endif
+
+#if PRAGMA_PACK
+#pragma pack(1)
+#endif
+
+/*****************************************************************************
+ * Module descriptor
+ *****************************************************************************/
+#define CACHING_TEXT N_("Caching value in ms")
+#define CACHING_LONGTEXT N_( \
+    "Caching value for Truran VOD streams. This " \
+    "value should be set in milliseconds." )
+
+#define RECONNECT_TEXT N_("Auto re-connect")
+#define RECONNECT_LONGTEXT N_( \
+    "Automatically try to reconnect to the stream in case of a sudden " \
+    "disconnect." )
+
+static int  Open ( vlc_object_t * );
+static void Close( vlc_object_t * );
+
+vlc_module_begin();
+    set_shortname( N_("VOD" ) );
+    set_description( N_("Truran VOD input") );
+    set_category( CAT_INPUT );
+    set_subcategory( SUBCAT_INPUT_ACCESS );
+
+    add_integer( "vod-caching", DEFAULT_PTS_DELAY / 1000, NULL, CACHING_TEXT,
+                 CACHING_LONGTEXT, true );
+
+     add_bool( "vod-reconnect", 0, NULL, RECONNECT_TEXT,
+              RECONNECT_LONGTEXT, true );
+
+    set_capability( "access", 0 );
+    add_shortcut( "vod" );
+
+    set_callbacks( Open, Close );
+vlc_module_end();
+
+/*****************************************************************************
+ * Local prototypes
+ *****************************************************************************/
+#if PRAGMA_PACK
+#pragma pack(1)
+#endif
+
+typedef struct {
+    char magic[4];
+    uint32_t ver;		/* always 00 02 00 00 */
+    uint32_t type;		/* ok = 00 00 00 00, err = 0c 00 00 00 */
+    uint32_t unknown1;	/* always 00 00 00 00 */
+    uint32_t len_lo;		/* lower 32 bits of length */
+    uint32_t len_hi;		/* higher 32 bits of length */
+    uint32_t pos_lo;		/* lower 32 bits of position */
+    uint32_t pos_hi;		/* higher 32 bits of position */
+    uint32_t unknown2;	/* always 00 00 01 00 */
+    uint32_t unknown3;	/* always 00 00 00 00 */
+    uint32_t unknown4;	/* ok = 64 00 00 00, err = 00 00 00 00 */
+    char errmsg[100];
+    char name[260];
+    char date[108];
+} ATTRIBUTE_PACKED response_t;
+
+#if PRAGMA_PACK
+#pragma pack()
+#endif
+
+struct access_sys_t {
+    int fd;
+
+    /* From uri */
+    vlc_url_t url;
+
+    int i_code; /*must be 200*/
+    char       *psz_mime;
+    response_t resp;
+
+    bool b_reconnect;
+};
+
+/* */
+static ssize_t Read( access_t *, uint8_t *, size_t );
+
+#ifdef TRURAN_BLOCK_ACCESS
+static block_t *Block( access_t *p_access );
+#endif
+
+static int Seek( access_t *, int64_t );
+static int Control( access_t *, int, va_list );
+
+/* */
+static int Connect( access_t *, int64_t );
+static int Request( access_t *p_access, int64_t i_tell );
+static void Disconnect( access_t * );
+
+/*****************************************************************************
+ * Open:
+ *****************************************************************************/
+static int Open( vlc_object_t *p_this )
+{
+    access_t     *p_access = (access_t*)p_this;
+    access_sys_t *p_sys;
+
+    /* Set up p_access */
+#ifndef TRURAN_BLOCK_ACCESS
+    STANDARD_READ_ACCESS_INIT;
+#else
+    STANDARD_BLOCK_ACCESS_INIT;
+#endif
+
+    p_sys->fd = -1;
+    p_sys->psz_mime = NULL;
+
+    p_access->info.i_size = -1;
+    p_access->info.i_pos  = 0;
+    p_access->info.b_eof  = false;
+
+    /* Parse URI - remove spaces */
+    vlc_UrlParse( &p_sys->url,  p_access->psz_path, 0 );
+
+    if( p_sys->url.psz_host == NULL || *p_sys->url.psz_host == '\0' )
+    {
+        msg_Warn( p_access, "invalid host" );
+        goto error;
+    }
+
+    if( p_sys->url.i_port <= 0 )  p_sys->url.i_port = 1680;
+
+    p_sys->b_reconnect = var_CreateGetBool( p_access, "vod-reconnect" );
+
+    if (Connect( p_access, 0 ) != VLC_SUCCESS) goto error;
+
+    return VLC_SUCCESS;
+
+error:
+    vlc_UrlClean( &p_sys->url );
+    free( p_sys->psz_mime );
+
+    Disconnect( p_access );
+    free( p_sys );
+
+    return VLC_EGENERIC;
+}
+
+/*****************************************************************************
+ * Close:
+ *****************************************************************************/
+static void Close( vlc_object_t *p_this )
+{
+    access_t     *p_access = (access_t*)p_this;
+    access_sys_t *p_sys = p_access->p_sys;
+
+    vlc_UrlClean( &p_sys->url );
+    free( p_sys->psz_mime );
+
+    Disconnect( p_access );
+
+    free( p_sys );
+}
+
+/*****************************************************************************
+ * Read: Read up to i_len bytes from the http connection and place in
+ * p_buffer. Return the actual number of bytes read
+ *****************************************************************************/
+static ssize_t Read( access_t *p_access, uint8_t *p_buffer, size_t i_len )
+{
+    access_sys_t *p_sys = p_access->p_sys;
+    int i_read;
+
+    if( p_sys->fd < 0 )
+    {
+        p_access->info.b_eof = true;
+        return 0;
+    }
+
+    if( p_access->info.i_size >= 0 &&
+        i_len + p_access->info.i_pos > p_access->info.i_size )
+    {
+        if( ( i_len = p_access->info.i_size - p_access->info.i_pos ) == 0 )
+        {
+            p_access->info.b_eof = true;
+            return 0;
+        }
+    }
+
+    i_read = net_Read( p_access, p_sys->fd, NULL, p_buffer, i_len, false );
+
+#ifdef VOD_DEBUG
+    msg_Dbg(p_access, ">%d--->%d", i_read, i_len - i_read);
+#endif
+
+    if( i_read > 0 )
+    {
+        p_access->info.i_pos += i_read;
+    }
+    else if( i_read == 0 )
+    {
+        Disconnect( p_access );
+        if( p_sys->b_reconnect )
+        {
+            msg_Dbg( p_access, "got disconnected, trying to reconnect" );
+            if( Connect( p_access, p_access->info.i_pos ) )
+            {
+                msg_Dbg( p_access, "reconnection failed" );
+            }
+            else
+            {
+                p_sys->b_reconnect = false;
+                i_read = Read( p_access, p_buffer, i_len );
+                p_sys->b_reconnect = true;
+            }
+        }
+
+        if( i_read == 0 ) p_access->info.b_eof = true;
+    }
+
+    return i_read;
+}
+
+#ifdef TRURAN_BLOCK_ACCESS
+/*****************************************************************************
+ * Block
+ *****************************************************************************/
+static block_t *Block( access_t *p_access )
+{
+    access_sys_t *p_sys = p_access->p_sys;
+    block_t      *p_block;
+    ssize_t i_read;
+
+    if( p_access->info.b_eof )
+        return NULL;
+
+    /* Read data */
+    p_block = block_New( p_access, MTU );
+    i_read = net_Read( p_access, p_sys->fd, NULL, p_block->p_buffer, MTU, false );
+
+#ifdef VOD_DEBUG
+     msg_Dbg(p_access, "<%d--->%d", i_read, MTU - i_read);
+#endif
+
+    if( i_read < 0 )
+    {
+        block_Release( p_block );
+        return NULL;
+    }
+
+    return block_Realloc( p_block, 0, p_block->i_buffer = i_read );
+}
+#endif
+
+#if 0
+static char* get_comment(char* fn)
+{
+    char* pos;
+
+    pos = strrchr(fn,'\\');
+    if (pos) fn = pos+1;
+    pos = strrchr(fn,'/');
+    if (pos) fn = pos+1;
+    return fn;
+    //  return "comic.sjtu.edu.cn";
+}
+#endif
+
+static uint32_t get_ip_from_fd(int fd)
+{
+    struct sockaddr_in saddr;
+    socklen_t len = sizeof(saddr);
+
+    getsockname(fd,(struct sockaddr*)(&saddr), &len);
+    return saddr.sin_addr.s_addr;
+}
+
+static int Connect( access_t *p_access, int64_t i_tell )
+{
+    access_sys_t   *p_sys = p_access->p_sys;
+
+    /* Open connection */
+    p_sys->fd = net_ConnectTCP( p_access,  p_sys->url.psz_host,  p_sys->url.i_port );
+    if( p_sys->fd == -1 )
+    {
+        msg_Err( p_access, "cannot connect to %s:%d",  p_sys->url.psz_host,  p_sys->url.i_port );
+        return VLC_EGENERIC;
+    }
+    return Request( p_access, i_tell );
+}
+
+static int Request( access_t *p_access, int64_t i_tell )
+{
+    access_sys_t   *p_sys = p_access->p_sys;
+    char buffer[4096];
+    int pos = 0;
+    int size = sizeof(buffer);
+    response_t *resp = &p_sys->resp;
+    char           *psz;
+    char hostname[256];
+    gethostname (hostname, sizeof (hostname));
+
+    #define myprintf(fmt, args...) do { int len = snprintf(buffer+pos, size, fmt, ## args); pos+=len; size-=len; } while (0)
+
+    myprintf("POST /XMMS/%s%ld.xms HTTP/1.0\r\n", hostname, time(NULL) % 10000 + 10000);
+    myprintf("Host: %s\r\n",   hostname);
+    myprintf("Accept: %s\r\n", "*/*");
+    myprintf("User-Agent: %s\r\n", "XMPLAYER");
+    myprintf("UGT:%s\r\n",     "XMPLAYER");
+    myprintf("UIP:%"PRIu32"\r\n",     get_ip_from_fd(p_sys->fd));
+    myprintf("USN:%d\r\n",     12345678);
+    myprintf("PRI:%d\r\n",     0);
+    myprintf("UPR:%d\r\n",     0);
+    myprintf("SPL:%"PRId64"\r\n",     i_tell);
+    myprintf("SPH:%d\r\n",     0);
+    myprintf("URN:%s\r\n",     "sjtu");
+    myprintf("PWD:%s\r\n",     "sjtu");
+    myprintf("FPT:%d\r\n",     (i_tell==0)?1:0);
+    myprintf("FIL:%s\r\n",     (p_sys->url.psz_path[0] == '/')? p_sys->url.psz_path+1:p_sys->url.psz_path);
+    myprintf("ULG:%s\r\n",     "CN");
+    myprintf("STT:%s\r\n",     "YYYY/MM/DD HH:MM:SS");
+    myprintf("HST:%s\r\n",     hostname);
+    myprintf("PNM:%s\r\n",     "comic.sjtu.edu.cn");
+    myprintf("PID:%d\r\n",     0);
+    myprintf("IDX:%d\r\n",     0);
+    myprintf("\r\n");
+
+    msg_Dbg( p_access, "sending request: \n%s", buffer);
+
+    if ( net_Write( p_access, p_sys->fd, NULL, buffer, strlen(buffer) ) <= 0)
+    {
+        msg_Err( p_access, "failed to send request" );
+        Disconnect( p_access );
+        return VLC_EGENERIC;
+    }
+
+    /* Read Answer */
+    if( ( psz = net_Gets( VLC_OBJECT(p_access), p_sys->fd, NULL ) ) == NULL )
+    {
+        msg_Err( p_access, "failed to read answer" );
+        goto error;
+    }
+    if( !strncmp( psz, "HTTP/1.", 7 ) )
+    {
+          p_sys->i_code = atoi( &psz[9] );
+    }
+    else {
+        msg_Err( p_access, "invalid HTTP reply '%s'", psz );
+        free( psz );
+        goto error;
+    }
+
+    msg_Dbg( p_access, "answer code %d",  p_sys->i_code );
+
+    for( ;; )
+    {
+        char *psz = net_Gets( VLC_OBJECT(p_access), p_sys->fd, NULL );
+        char *p;
+
+        if( psz == NULL )
+        {
+            msg_Err( p_access, "failed to read answer" );
+            goto error;
+        }
+
+        if( !vlc_object_alive (p_access) || p_access->b_error )
+        {
+            free( psz );
+            goto error;
+        }
+
+        msg_Dbg( p_access, "Line=%s", psz );
+        if( *psz == '\0' )
+        {
+            free( psz );
+            break;
+        }
+
+        if( ( p = strchr( psz, ':' ) ) == NULL )
+        {
+            msg_Err( p_access, "malformed header line: %s", psz );
+            free( psz );
+            goto error;
+        }
+        *p++ = '\0';
+        while( *p == ' ' ) p++;
+
+        if( !strcasecmp( psz, "Content-Type" ) )
+        {
+            free( p_sys->psz_mime );
+            p_sys->psz_mime = strdup( p );
+            msg_Dbg( p_access, "Content-Type: %s", p_sys->psz_mime );
+            if (!strcmp(p_sys->psz_mime, "errmsg/*"))
+            {
+                msg_Err( p_access, "Error: %s", p );
+                free( psz );
+                goto error;
+            }
+        }
+
+        free( psz );
+    }
+
+    if (net_Read( p_access, p_sys->fd, NULL, resp, sizeof(response_t), true ) < (int)sizeof(response_t))
+    {
+        msg_Err( p_access, "failed to read Response" );
+        goto error;
+    }
+
+    if (strncmp( resp->magic, "XMMS", 4 ) != 0)
+    {     // CHECK MAGIC
+        msg_Err(p_access, " magic mismatch");
+        goto error;
+    }
+
+    if ((resp->type != 0) && (resp->unknown4 == 0))
+    { // maybe error
+        msg_Err(p_access, "Server error %s", resp->errmsg);
+        goto error;
+    }
+
+    msg_Dbg(p_access, "Opening %s\n", resp->name);
+    msg_Dbg(p_access, "Date %s\n", resp->date);
+
+#if defined( WORDS_BIGENDIAN )
+            p_access->info.i_size = resp->len_lo >> 24 + (resp->len_lo &  0x00ff0000) >> 8 +
+                                                                                            (  resp->len_lo & 0x0000ff00 ) << 8 +  resp->len_lo <<  24;
+            p_access->info.i_pos  = resp->pos_lo >> 24 + (resp->pos_lo &  0x00ff0000) >> 8 +
+                                                                                            (  resp->pos_lo & 0x0000ff00 ) << 8 +  resp->pos_lo <<  24;
+#else
+            p_access->info.i_size = resp->len_lo;
+            p_access->info.i_pos  =resp->pos_lo;
+#endif
+
+    msg_Dbg(p_access, "Position %"PRId64"/%"PRId64"\n", p_access->info.i_pos,  p_access->info.i_size);
+
+    if (resp->len_lo == 0) {
+        Disconnect(p_access);
+        msg_Err( p_access, "Open failed, may be too many users." );
+        return VLC_EGENERIC;
+    }
+
+    p_access->info.b_eof  = false;
+
+    if ( p_access->info.i_pos >= p_access->info.i_size )
+        p_access->info.b_eof  = true;
+
+    return VLC_SUCCESS;
+
+error:
+    Disconnect( p_access );
+    return VLC_EGENERIC;
+}
+
+/*****************************************************************************
+ * Disconnect:
+ *****************************************************************************/
+static void Disconnect( access_t *p_access )
+{
+    access_sys_t *p_sys = p_access->p_sys;
+
+    if( p_sys->fd != -1)
+    {
+        net_Close(p_sys->fd);
+        p_sys->fd = -1;
+    }
+}
+
+/*****************************************************************************
+ * Seek: close and re-open a connection at the right place
+ *****************************************************************************/
+static int Seek( access_t *p_access, int64_t i_pos )
+{
+    msg_Dbg( p_access, "trying to seek to %"PRId64, i_pos );
+
+    Disconnect( p_access );
+
+    if( p_access->info.i_size
+     && (uint64_t)i_pos >= (uint64_t)p_access->info.i_size ) {
+        msg_Err( p_access, "seek to far" );
+        int retval = Seek( p_access, p_access->info.i_size - 1 );
+        if( retval == VLC_SUCCESS ) {
+            uint8_t p_buffer[2];
+            Read( p_access, p_buffer, 1);
+            p_access->info.b_eof  = false;
+        }
+        return retval;
+    }
+    if( Connect( p_access, i_pos ) != VLC_SUCCESS)
+    {
+        msg_Err( p_access, "seek failed" );
+        p_access->info.b_eof = true;
+        return VLC_EGENERIC;
+    }
+    return VLC_SUCCESS;
+}
+
+/*****************************************************************************
+ * Control:
+ *****************************************************************************/
+static int Control( access_t *p_access, int i_query, va_list args )
+{
+    access_sys_t *p_sys = p_access->p_sys;
+    bool   *pb_bool;
+    int64_t      *pi_64;
+
+    switch( i_query )
+    {
+        /* */
+        case ACCESS_CAN_SEEK:
+            pb_bool = (bool*)va_arg( args, bool* );
+            *pb_bool =true;
+            break;
+        case ACCESS_CAN_FASTSEEK:
+            pb_bool = (bool*)va_arg( args, bool* );
+            *pb_bool = false;
+            break;
+        case ACCESS_CAN_PAUSE:
+        case ACCESS_CAN_CONTROL_PACE:
+            pb_bool = (bool*)va_arg( args, bool* );
+
+            *pb_bool = true;
+            break;
+
+        case ACCESS_GET_PTS_DELAY:
+            pi_64 = (int64_t*)va_arg( args, int64_t * );
+            *pi_64 = (int64_t)var_GetInteger( p_access, "vod-caching" ) * INT64_C(1000);
+            break;
+
+        /* */
+        case ACCESS_SET_PAUSE_STATE:
+            break;
+
+        case ACCESS_GET_CONTENT_TYPE:
+            *va_arg( args, char ** ) =
+                p_sys->psz_mime ? strdup( p_sys->psz_mime ) : NULL;
+            break;
+
+        case ACCESS_GET_TITLE_INFO:
+        case ACCESS_SET_TITLE:
+        case ACCESS_SET_SEEKPOINT:
+        case ACCESS_SET_PRIVATE_ID_STATE:
+            return VLC_EGENERIC;
+
+        default:
+            msg_Warn( p_access, "unimplemented query in control" );
+            return VLC_EGENERIC;
+
+    }
+    return VLC_SUCCESS;
+}
+
-- 
1.5.4.3

